class InvaildCredentials(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

class DrivePathException(FileNotFoundError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


import io
from tkinter import NO
from typing import Union
from googleapiclient.http import MediaIoBaseDownload, MediaFileUpload, MediaIoBaseUpload
from googleapiclient.errors import HttpError
from .Exceptions import DrivePathException
from googleapiclient.discovery import Resource
class Folder:
    def __init__(self, folder_name:str, folder_id:str, service:Resource, parent:'Folder'):
        """
        Initialize a Folder instance.

        Args:
            folder_name (str): The name of the folder.
            folder_id (str): The unique ID of the folder.
            service (googleapiclient.discovery.Resource): The Google Drive service instance.
            parent (Folder):the parent Folder
        """
        self.__folder_name = folder_name
        self.__id = folder_id
        self.__service = service
        self.__parent = parent
    def get_id(self):
        return self.__id
    def getParent(self)->'Folder':
        return self.__parent
    def name(self):return self.__folder_name
    def get_files(self)->list['File']:
        """
        Get all files in the folder.

        Returns:
            list[File]: List of File objects in the folder.
        """
        query = f"'{self.__id}' in parents and mimeType!='application/vnd.google-apps.folder'"
        try:
            files=[]
            results = self.__service.files().list(q=query, fields="files(id, name)").execute() # type: ignore
            items = results.get('files', [])
            for item in items:
                item=dict(item)
                if not item.get('mimeType') or not item['mimeType'].startswith('application/vnd.google-apps.'):
                        file = File(item['name'], item['id'], self.__service)
                else: 
                        file = Doc(item['name'], item['id'], self.__service)
                files.append(file)
            return files
        except HttpError as error:
            print(f"An error occurred: {error}")
            return []
    def get_folders(self)->list['Folder']:
        """
        Get all subfolders in the folder.

        Returns:
            list[Folder]: List of Folder objects in the folder.
        """
        query = f"'{self.__id}' in parents and mimeType='application/vnd.google-apps.folder'"
        try:
            results = self.__service.files().list(q=query, fields="files(id, name)").execute() # type: ignore
            items = results.get('files', [])
            return [Folder(item['name'], item['id'], self.__service,self) for item in items]
        except HttpError as error:
            print(f"An error occurred: {error}")
            return []
    def get_folder_by_name(self, name:str)->Union['Folder',None]:
        """
        Gets a folder by name
        Args:
            name(str)
        Returns:
            Folder or None if A folder was found return the folder
        """
        for folder in self.get_folders():
            if folder.name() == name:
                return folder
        return folder
    
    def get_file_by_name(self, name:str)->Union['File',None]:
        """
        Gets a file by name
        Args:
            name(str)
        Returns:
            File or None if A File was found return the File if not return None
        """
        for file in self.get_files():
            if file.name() == name:
                return file
        return None
    def rename(self, new_foldername):
        """
        Rename the folder.

        Args:
            new_foldername (str): The new name for the folder.
        """
        try:
            file_metadata = {'name': new_foldername}
            self.__service.files().update(fileId=self.__id, body=file_metadata).execute() # type: ignore
            self.__folder_name = new_foldername
        except HttpError as error:
            print(f"An error occurred: {error}")

    def delete(self):
        """
        Delete the folder from Google Drive.

        Raises:
            HttpError: An error occurred while deleting the folder.
        """
        try:
            self.__service.files().delete(fileId=self.__id).execute() # type: ignore
        except HttpError as error:
            raise DrivePathException(f"An error occurred: {error}")
    
    def __str__(self):
        return self.__folder_name

    def __repr__(self) -> str:
        return str(self)


class File:
    def __init__(self, filename:str, file_id:str, service:Resource):
        """
        Initialize a File instance.

        Args:
            filename (str): The name of the file.
            file_id (str): The unique ID of the file.
            service (googleapiclient.discovery.Resource): The Google Drive service instance.
        """
        self.__filename = filename
        self.__id = file_id
        self.__service = service
    def name(self):return self.__filename
    def rename(self, new_filename):
        """
        Rename the file.

        Args:
            new_filename (str): The new name for the file.
        """
        try:
            file_metadata = {'name': new_filename}
            self.__service.files().update(fileId=self.__id, body=file_metadata).execute() # type: ignore
            self.__filename = new_filename
        except HttpError as error:
            print(f"An error occurred: {error}")

    def delete(self):
        """
        Delete the file from Google Drive.
        """
        try:
            self.__service.files().delete(fileId=self.__id).execute() # type: ignore
        except HttpError as error:
            print(f"An error occurred: {error}")

    def read(self) -> bytes:
        """
        Read the contents of the file.

        Returns:
            bytes: The contents of the file.
        """
        try:
            request = self.__service.files().get_media(fileId=self.__id) # type: ignore
            fh = io.BytesIO()
            downloader = MediaIoBaseDownload(fh, request)
            done = False
            while not done:
                status, done = downloader.next_chunk()
            fh.seek(0)
            return fh.read()
        except HttpError as error:
            print(f"An error occurred: {error}")
            return b''

    def write(self, content: bytes, mime_type='application/octet-stream'):
        """
        Write contents to the file.

        Args:
            content (bytes): The content to write to the file.
            mime_type (str): The MIME type of the file content.
        """
        try:
            media = MediaIoBaseUpload(io.BytesIO(content), mimetype=mime_type)
            self.__service.files().update(fileId=self.__id, media_body=media).execute() # type: ignore
        except HttpError as error:
            print(f"An error occurred: {error}")

    def __str__(self) -> str:
        return self.__filename

    def __repr__(self) -> str:
        return str(self)
class Doc(File):
    def __init__(self, filename, file_id, service):
        """
        Initialize a Doc instance.

        Args:
            filename (str): The name of the document.
            file_id (str): The unique ID of the document.
            service (googleapiclient.discovery.Resource): The Google Drive service instance.
        """
        super().__init__(filename, file_id, service)
        self.__service=service
        self.__id=file_id
    def rename(self, new_filename):
        """
        Rename the document.

        Args:
            new_filename (str): The new name for the document.
        """
        super().rename(new_filename)

    def delete(self):
        """
        Delete the document from Google Drive.
        """
        super().delete()

    def read(self) -> bytes:
        """
        Read the contents of the document and export it as a DOCX file.

        Returns:
            bytes: The contents of the document as a DOCX file.
        """
        try:
            request = self.__service.files().export_media(fileId=self.__id, mimeType='application/vnd.openxmlformats-officedocument.wordprocessingml.document') # type: ignore
            fh = io.BytesIO()
            downloader = MediaIoBaseDownload(fh, request)
            done = False
            while not done:
                status, done = downloader.next_chunk()
            fh.seek(0)
            return fh.read()
        except HttpError as error:
            print(f"An error occurred: {error}")
            return b''
    def write(self, content: bytes):
        """
        Write contents to the document.

        Args:
            content (bytes): The content(DOCX file like bytes) to write to the document.
        """
        super().write(content, mime_type='application/vnd.openxmlformats-officedocument.wordprocessingml.document')

class RootFolder(Folder):
    def __init__(self,service:Resource)->None:
       self.__service = service
    def __build(self):
         
        """
            Builds the initial list of folders and files.
        """
        files = []
        folders = []
        query = "mimeType='application/vnd.google-apps.folder'"
        results = self.__service.files().list(q=query, fields="nextPageToken, files(id, name)").execute() # type: ignore
        items = results.get('files', [])
        
        for item in items:
            folder = Folder(item['name'], item['id'], self.__service,self)
            folders.append(folder)

        query = (
            "mimeType!='application/vnd.google-apps.folder' and "
            "mimeType!='application/vnd.google-apps.spreadsheet' and "
            "mimeType!='application/vnd.google-apps.presentation' and "
            "mimeType!='application/vnd.google-apps.form' and "
            "mimeType!='application/vnd.google-apps.drawing' and "
            "mimeType!='application/vnd.google-apps.map' and "
            "mimeType!='application/vnd.google-apps.script' and "
            "mimeType!='application/vnd.google-apps.site' and "
            "mimeType!='application/vnd.google-apps.jam' and "
            "mimeType!='application/vnd.google-apps.shortcut'"
        )
        results = self.__service.files().list(q=query, fields="nextPageToken, files(id, name, mimeType)").execute() # type: ignore
        items = results.get('files', [])

        for item in items:
            if item['mimeType'].startswith('application/vnd.google-apps.'):
                file = Doc(item['name'], item['id'], self.__service)
            else:
                file = File(item['name'], item['id'], self.__service)
            files.append(file)
        self.__folders=folders
        self.__files=files
    def get_folders(self)->list[Folder]:
        self.__build()
        return self.__folders
    def get_files(self) -> list['File']:
        self.__build()
        return self.__files
    def get_id(self):
        return None
    def delete(self):raise NotImplementedError("Can't delete the root folder")
    def rename(self, new_foldername):raise NotImplementedError("Can't rename the root folder")
    def name(self):return '/'
    def getParent(self):
        return self
    



from os import PathLike
import os.path
from typing import Union
from .Exceptions import InvaildCredentials, DrivePathException
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from .FileSystemTypes import *
from google.oauth2.credentials import Credentials

def login(tokenFile: str = "token.json", credentialsFile: str = "credentials.json") -> Credentials:
    """
    Logs in to the Google API and returns credentials.

    Args:
        tokenFile (PathLike): The path to the token file. Defaults to "token.json".
        credentialsFile (PathLike): The path to the credentials file. Defaults to "credentials.json".

    Returns:
        Credentials: The credentials object for the Google API.

    Raises:
        InvaildCredentials: If no valid credentials file is found.
    """
    creds = None
    SCOPES = [
        'https://www.googleapis.com/auth/drive.metadata.readonly',
        'https://www.googleapis.com/auth/docs',
        'https://www.googleapis.com/auth/drive',
        'https://www.googleapis.com/auth/drive.appdata',
        'https://www.googleapis.com/auth/drive.file'
    ]
    if os.path.exists(tokenFile):
        creds = Credentials.from_authorized_user_file(tokenFile, SCOPES)
    if not creds or not creds.valid:
        if not os.path.exists(credentialsFile):
            raise InvaildCredentials("No credentials file found or the path was invalid")
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(credentialsFile, SCOPES)
            creds = flow.run_local_server(port=0)
            with open(tokenFile, "w") as token:
                token.write(creds.to_json())
    return creds # type: ignore

class Drive:
    def __init__(self, creds) -> None:
        """
        Initializes a Google Drive instance.

        Args:
            creds: Credentials object for Google Drive API.
        """
        self.__service = build("drive", "v3", credentials=creds)
        # self.__currentDir = "/"
        
        self.__root: RootFolder = RootFolder(self.__service)
        self.__currentFolder:Folder = self.__root
        self.__currentFolder = self.__root
    def cd(self, path: str) -> None:
        """
        Changes the current directory.

        Args:
            path (str): The path to change to.

        Raises:
            DrivePathException: If the path is invalid.
        """
        
        folder = self.__transverse_path(path)
        if not isinstance(folder, Folder):
            raise DrivePathException(f"No such directory as {folder}")
        self.__currentFolder = folder
        if path in '..':path = path.replace("..", self.__getParent().name())

    def __transverse_path(self, path: str) -> Folder:
        """
        Transverses the path to find the target folder .

        Args:
            path (str): The path to transverse.

        Returns:
            Folder object at the end of the path.

        Raises:
            DrivePathException: If the path is invalid.
        """
        
        parts = path.split("/") 
        if path.startswith("/"):
            current_folder = self.__root
        else:
            current_folder = self.__currentFolder if not path.startswith("..") else self.__currentFolder.getParent()
        if parts[0] == "" or parts[0] == "..":
            parts.pop(0)
        for part in parts:
            if current_folder is not None:
                current_folder = current_folder.get_folder_by_name(part)
            if current_folder is None:
                raise DrivePathException(f"No such directory {part}")
        if current_folder is None:
            raise DrivePathException(f"Invalid path: {path}")
        return current_folder

    def __getParent(self) -> Folder:
        """
        Retrieves the parent folder of the current directory.

        Returns:
            Parent Folder object or None if in root.
        """
        
        if self.__currentFolder.name() != "/":
            parts = self.getCurrentFolder().split("/")
            parts.pop()
            parts.pop()
            if len(parts) > 1: return self.__root
            parent_path = "/".join(parts)
            return self.__transverse_path(parent_path)
        else:
            return self.__root

    def dir(self, path: str = None) -> list[Union[File, Folder]]: # type: ignore
        """
        Lists the contents of the given path or the current folder.

        Args:
            path (str, optional): The path to list contents of. Defaults to None.

        Returns:
            list of File and Folder objects in the specified path or current folder.

        Raises:
            DrivePathException: If the path is invalid.
        """
        
        if path:
            target = self.__transverse_path(path)
        else:
            target = self.__currentFolder

        if isinstance(target, Folder) or issubclass(type(target),Folder):
            files, folders = target.get_files(), target.get_folders()
        else:
            raise DrivePathException(f"No such directory as {path}")

        return files + folders
    
    def file(self, path: str,type='text/plain') -> File:
        """
        get's a file
        Creates a file at the given path if one does Not exist.

        Args:
            path (str): The path where the file should be created.
            type(str, optional): The file type that you want to use to create a folder if none exist defaults to plain text

        Returns:
            the file
        """
        
        path = path.replace("\\", "/")
        parts = path.split("/")
        filename = parts.pop()

        if parts:
            folder_path = "/".join(parts)
            folder = self.__transverse_path(folder_path)
            if not isinstance(folder, Folder):
                raise DrivePathException(f"No such directory: {folder_path}")
        else:
            folder = self.__currentFolder

        # Check if the file already exists
        try:
            existing_file = next((f for f in folder.get_files() if f.__str__() == filename), None)
        except TypeError: existing_file=False
        if existing_file:
            return folder.get_file_by_name(filename) # type: ignore

        # Create the file
        file_metadata = {
            'name': filename,
            'parents': [folder.get_id()],
            'mimeType':type
        }
        self.__service.files().create(body=file_metadata, fields='id').execute()
        return folder.get_file_by_name(filename) # type: ignore
    def mkdir(self, path: str) -> None:
        """
        Creates a folder at the given path if one does not already exist.

        Args:
            path (str): The path where the folder should be created.

        Raises:
            DrivePathException: If the path is invalid or the folder already exists.
        """
        
        path = path.replace("\\", "/")
        parts = path.split("/")
        folder_name = parts.pop()

        if parts:
            parent_folder_path = "/".join(parts)
            parent_folder = self.__transverse_path(parent_folder_path)
        else:
            parent_folder = self.__currentFolder if self.__currentFolder else self.__root

        # Check if the folder already exists
        existing_folder = next((f for f in parent_folder.get_folders() if f.__str__() == folder_name), None)
        if existing_folder:
            raise DrivePathException(f"Folder '{folder_name}' already exists at path '{path}'.")

        # Create the folder
        folder_metadata = {
            'name': folder_name,
            'mimeType': 'application/vnd.google-apps.folder',
            'parents': [parent_folder.get_id()]
        }
        self.__service.files().create(body=folder_metadata, fields='id').execute()
    def rmdir(self, path: str) -> None:
            
        path = path.replace("\\", "/")
        parts = path.split("/")
        folder_name = parts.pop()

        if parts:
            parent_folder_path = "/".join(parts)
            parent_folder = self.__transverse_path(parent_folder_path)
        else:
            parent_folder = self.__currentFolder if self.__currentFolder else self.__root
        folder=parent_folder.get_folder_by_name(folder_name)
        if not folder:raise DrivePathException(f"Folder '{folder_name}' does not exist at path '{path}'")
        folder.delete()
    def getCurrentFolder(self)->str:
        """
        Gets the path to the current folder
        Returns:
            the path to the current folder
        """
        folders:list[Folder]=[self.__currentFolder]
        parent=self.__currentFolder.getParent()
        if parent is not None and not isinstance(parent, RootFolder):
            folders.append(parent)
        while parent is not None and not isinstance(parent, RootFolder):
            parent=parent.getParent()
            if parent is not None and not isinstance(parent,RootFolder):
                folders.append(parent)
        folders.reverse()
        path=""
        for folder in folders:
            path+="/"
            if isinstance(folder,RootFolder):continue
            path+=folder.name()

        return path
            


# Common File MIME Types
PDF_MIME = 'application/pdf'
TEXT_MIME = 'text/plain'
HTML_MIME = 'text/html'
CSV_MIME = 'text/csv'

# Google Workspace MIME Types
GOOGLE_DOC_MIME = 'application/vnd.google-apps.document'
# Other google workspace types coming soon

# GOOGLE_SHEET_MIME = 'application/vnd.google-apps.spreadsheet'
# GOOGLE_SLIDE_MIME = 'application/vnd.google-apps.presentation'
# GOOGLE_FORM_MIME = 'application/vnd.google-apps.form'
# GOOGLE_DRAWING_MIME = 'application/vnd.google-apps.drawing'
# GOOGLE_MAP_MIME = 'application/vnd.google-apps.map'
# GOOGLE_SCRIPT_MIME = 'application/vnd.google-apps.script'
# GOOGLE_SITE_MIME = 'application/vnd.google-apps.site'
# GOOGLE_JAMBOARD_MIME = 'application/vnd.google-apps.jam'
# GOOGLE_SHORTCUT_MIME = 'application/vnd.google-apps.shortcut'
# GOOGLE_FOLDER_MIME = 'application/vnd.google-apps.folder'
# IMAGE_JPEG_MIME = 'image/jpeg'
# IMAGE_PNG_MIME = 'image/png'
# IMAGE_GIF_MIME = 'image/gif'
# AUDIO_MP3_MIME = 'audio/mpeg'
# VIDEO_MP4_MIME = 'video/mp4'
# ZIP_MIME = 'application/zip'
# EPUB_MIME = 'application/epub+zip'
# WORD_MIME = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
# EXCEL_MIME = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
# POWERPOINT_MIME = 'application/vnd.openxmlformats-officedocument.presentationml.presentation'

# Other File MIME Types
JSON_MIME = 'application/json'
XML_MIME = 'application/xml'
RTF_MIME = 'application/rtf'
MARKDOWN_MIME = 'text/markdown'
MP3_MIME = 'audio/mpeg'
OGG_AUDIO_MIME = 'audio/ogg'
OGG_VIDEO_MIME = 'video/ogg'
MP4_VIDEO_MIME = 'video/mp4'
WEBM_VIDEO_MIME = 'video/webm'